use crate::parser::{satisfy, take, FailedParserResult, Parser};

pub fn char_parser<'a, TContext>(c: char) -> impl Parser<'a, char, char, TContext> {
    satisfy(move |x| *x == c)
}

pub fn string_parser<'a, TContext>(str: &'static str) -> impl Parser<'a, char, String, TContext> {
    take::<char, TContext>(str.chars().count()).next(move |result| {
        result.and_then(|(input, chars)| {
            let chars: String = chars.iter().collect();
            if chars == str {
                Ok((input, chars))
            } else {
                Err((
                    input,
                    FailedParserResult::new(format!("Failed to parse '{}'.", str)),
                ))
            }
        })
    })
}

pub fn one_of<'a, TContext>(str: &'static str) -> impl Parser<'a, char, char, TContext> {
    let str: Vec<char> = str.chars().collect();
    satisfy(move |c: &char| str.contains(c))
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::parser::{
        failed_parser_test_helper, parser_test_helper, ParserContext, ParserResult,
    };
    use std::cell::RefCell;
    use std::rc::Rc;

    #[test]
    fn char_parser_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'a',
            |context, input| char_parser('a').parse(context, input),
        );

        failed_parser_test_helper(ParserContext::new_with_str("bc", ()), |context, input| {
            char_parser('a').parse(context, input)
        });
    }

    #[test]
    fn string_parser_test() {
        parser_test_helper(
            ParserContext::new_with_str("Hello, world!", ()),
            &"Hello".to_owned(),
            |context, input| string_parser("Hello").parse(context, input),
        );

        fn test_parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, String> {
            let (input, first) = string_parser("hello, ").parse(context.clone(), input)?;
            let (input, second) = string_parser("world!").parse(context.clone(), input)?;

            Ok((input, first + second.as_str()))
        }

        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &"hello, world!".to_owned(),
            test_parser,
        );
        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &(),
            |context, input| test_parser.map(|_| ()).parse(context, input),
        )
    }
}
